// -*- C++ -*-
#include "Rivet/Analysis.hh"
#include "Rivet/RivetAIDA.hh"

#include "Rivet/Projections/ZFinder.hh"
#include "Rivet/Projections/FastJets.hh"
#include "Rivet/Projections/VetoedFinalState.hh"

#include<iostream>


namespace Rivet {


  class ATLAS_2013_I1230812 : public Analysis {
  public:

    /// @name Constructors etc.
    //@{

    /// Constructor
    ATLAS_2013_I1230812(std::string name="ATLAS_2013_I1230812")
      : Analysis(name),
        _weights_incl(7, 0.0),
        _weights_excl(7, 0.0),
        _weights_excl_pt150(7, 0.0),
        _weights_excl_vbf(7, 0.0)
    {
      _mode = 1;
    }

    //@}


  public:

    /// Book histograms and initialise projections before the run
    void init() {

      if (_mode==1) {
        // combined
        ZFinder zfinder(-2.5, 2.5, 20, ELECTRON, 66.0*GeV, 116.0*GeV, 0.1, true, false);
        addProjection(zfinder, "zfinder");
      }
      else if (_mode==2) {
        // electron
        std::vector<std::pair<double, double> > eta_e;
        eta_e.push_back(make_pair(-2.47, -1.52));
        eta_e.push_back(make_pair(-1.37,  1.37));
        eta_e.push_back(make_pair( 1.52,  2.47));
        ZFinder zfinder(eta_e, 20, ELECTRON, 66.0*GeV, 116.0*GeV, 0.1, true, false);
        addProjection(zfinder, "zfinder");
      }
      else if (_mode==3) {
        // muon
        ZFinder zfinder(-2.4, 2.4, 20, MUON, 66.0*GeV, 116.0*GeV, 0.1, true, false);
        addProjection(zfinder, "zfinder");
      }

      // Define veto FS in order to prevent Z-decay products entering the jet algorithm
      VetoedFinalState had_fs;
      had_fs.addVetoOnThisFinalState(getProjection<ZFinder>("zfinder"));
      FastJets jets(had_fs, FastJets::ANTIKT, 0.4);
      jets.useInvisibles(true);
      addProjection(jets, "jets");

      _h_njet_incl  = bookHistogram1D (1,  1, _mode);
      _h_njet_incl_ratio = bookDataPointSet(2,  1, _mode);
      _h_njet_excl  = bookHistogram1D (3,  1, _mode);
      _h_njet_excl_ratio  = bookDataPointSet (4,  1, _mode);
      _h_njet_excl_pt150  = bookHistogram1D (5,  1, _mode);
      _h_njet_excl_pt150_ratio  = bookDataPointSet (6,  1, _mode);
      _h_njet_excl_vbf  = bookHistogram1D (7,  1, _mode);
      _h_njet_excl_vbf_ratio  = bookDataPointSet (8,  1, _mode);
      _h_ptlead      = bookHistogram1D (9,  1, _mode);
      _h_ptseclead  = bookHistogram1D (10,  1, _mode);
      _h_ptthirdlead  = bookHistogram1D (11,  1, _mode);
      _h_ptfourthlead  = bookHistogram1D (12,  1, _mode);
      _h_ptlead_excl      = bookHistogram1D (13,  1, _mode);
      _h_pt_ratio      = bookHistogram1D (14,  1, _mode);
      _h_pt_z      = bookHistogram1D (15,  1, _mode);
      _h_pt_z_excl      = bookHistogram1D (16,  1, _mode);
      _h_ylead      = bookHistogram1D (17,  1, _mode);
      _h_yseclead   = bookHistogram1D (18,  1, _mode);
      _h_ythirdlead   = bookHistogram1D (19,  1, _mode);
      _h_yfourthlead   = bookHistogram1D (20,  1, _mode);
      _h_deltay     = bookHistogram1D (21, 1, _mode);
      _h_mass       = bookHistogram1D (22,  1, _mode);
      _h_deltaphi   = bookHistogram1D (23, 1, _mode);
      _h_deltaR     = bookHistogram1D (24, 1, _mode);
      _h_ptthirdlead_vbf  = bookHistogram1D (25,  1, _mode);
      _h_ythirdlead_vbf   = bookHistogram1D (26,  1, _mode);
      _h_ht     = bookHistogram1D (27, 1, _mode);
      _h_st     = bookHistogram1D (28, 1, _mode);
    }


    /// Perform the per-event analysis
    void analyze(const Event& event) {

      const ZFinder& zfinder = applyProjection<ZFinder>(event, "zfinder");
      if (zfinder.constituents().size()!=2) vetoEvent;

      FourMomentum z = zfinder.bosons()[0].momentum();
      FourMomentum lp = zfinder.constituents()[0].momentum();
      FourMomentum lm = zfinder.constituents()[1].momentum();
      if (deltaR(lp, lm)<0.2) vetoEvent;

      Jets jets;
      /// @todo Replace with a Cut passed to jetsByPt
      foreach(const Jet& jet, applyProjection<FastJets>(event, "jets").jetsByPt(30.0*GeV)) {
        FourMomentum jmom = jet.momentum();
        if (fabs(jmom.rapidity()) < 4.4 && deltaR(lp, jmom) > 0.5  && deltaR(lm, jmom) > 0.5) {
          jets.push_back(jet);
        }
      }


      const double weight = event.weight();

      if (jets.size()<7) _weights_excl[jets.size()] += weight;
      for (size_t i=0; i<7; ++i) {
        if (jets.size()>=i) _weights_incl[i] += weight;
      }
      
      // Fill jet multiplicities
      for (size_t ijet=1; ijet<=jets.size(); ++ijet) {
	_h_njet_incl->fill(ijet, weight);
      }
      _h_njet_excl->fill(jets.size(), weight);

      // Require at least one jet
      if (jets.size() >= 1)
	{            
	  // Leading jet histos
	  const double ptlead   = jets[0].momentum().pT()/GeV;
	  const double yabslead = fabs(jets[0].momentum().rapidity());
	  const double ptz   = z.pT()/GeV;
	  _h_ptlead->fill(ptlead,   weight);
	  _h_ylead ->fill(yabslead, weight);
	  _h_pt_z  ->fill(ptz, weight);
	  // Fill jet multiplicities
	  if(ptlead>150)
	    {
	      _h_njet_excl_pt150->fill(jets.size(), weight);
              if (jets.size()<7) _weights_excl_pt150[jets.size()] += weight;
	    }
	  
	  // Loop over selected jets, fill inclusive distributions
	  double st=0;
	  double ht=lp.pT()/GeV+lm.pT()/GeV;
	  for (size_t ijet = 0; ijet < jets.size(); ++ijet) {
	    ht+=jets[ijet].momentum().pT()/GeV;
	    st+=jets[ijet].momentum().pT()/GeV;
	  }
	  _h_ht->fill(ht, weight);
	  _h_st->fill(st, weight);

	  // Require exactly one jet
	  if (jets.size() == 1)
	    {
	      _h_ptlead_excl->fill(ptlead,   weight);
	      _h_pt_z_excl  ->fill(ptz, weight);
	    }
	}
    

      // Require at least two jets
      if (jets.size() >= 2) {
	// Second jet histos
	const double ptlead      = jets[0].momentum().pT()/GeV;
	const double pt2ndlead   = jets[1].momentum().pT()/GeV;
	const double ptratio     = pt2ndlead/ptlead;
	const double yabs2ndlead = fabs(jets[1].momentum().rapidity());
	_h_ptseclead ->fill(pt2ndlead,   weight);
	_h_yseclead  ->fill(yabs2ndlead, weight);
	_h_pt_ratio  ->fill(ptratio, weight);
	
	// Dijet histos
	const double deltaphi = fabs(deltaPhi(jets[1], jets[0]));
	const double deltarap = fabs(jets[0].momentum().rapidity() - jets[1].momentum().rapidity()) ;
	const double deltar   = fabs(deltaR(jets[0], jets[1], RAPIDITY));
	const double mass     = (jets[0].momentum() + jets[1].momentum()).mass()/GeV;
	_h_mass     ->fill(mass,     weight);
	_h_deltay   ->fill(deltarap, weight);
	_h_deltaphi ->fill(deltaphi, weight);
	_h_deltaR   ->fill(deltar,   weight);

	if(mass>350&&deltarap>3.0)
	  {
	    _h_njet_excl_vbf->fill(jets.size(), weight);
            if (jets.size()<7) _weights_excl_vbf[jets.size()] += weight;
	  }
      }

      // Require at least three jets
      if (jets.size() >= 3) {
	// Third jet histos
	const double pt3rdlead   = jets[2].momentum().pT()/GeV;
	const double yabs3rdlead = fabs(jets[2].momentum().rapidity());
	_h_ptthirdlead ->fill(pt3rdlead,   weight);
	_h_ythirdlead  ->fill(yabs3rdlead, weight);

	//Histos after VBF preselection
	const double deltarap = fabs(jets[0].momentum().rapidity() - jets[1].momentum().rapidity()) ;
	const double mass     = (jets[0].momentum() + jets[1].momentum()).mass();
	if(mass>350&&deltarap>3.0)
	  {
	    _h_ptthirdlead_vbf ->fill(pt3rdlead,   weight);
	    _h_ythirdlead_vbf  ->fill(yabs3rdlead, weight);
	  }
      }

      // Require at least four jets
      if (jets.size() >= 4) {
	// Fourth jet histos
	const double pt4thlead   = jets[3].momentum().pT()/GeV;
	const double yabs4thlead = fabs(jets[3].momentum().rapidity());
	_h_ptfourthlead ->fill(pt4thlead,   weight);
	_h_yfourthlead  ->fill(yabs4thlead, weight);
      }
    }

    /// @name Ratio calculator util functions
    //@{
    
    /// Calculate the ratio, being careful about div-by-zero
    double ratio(double a, double b) {
      return (b != 0) ? a/b : 0;
    }

    /// Calculate the ratio error, being careful about div-by-zero
    double ratio_err_incl(double a, double b) {
      return (b != 0) ? sqrt(a/b*(1-a/b)/b) : 0;
    }

    /// Calculate the ratio error, being careful about div-by-zero
    double ratio_err_excl(double a, double b) {
      return (b != 0) ? sqrt(a/sqr(b) + sqr(a)/(b*b*b)) : 0;
    }

    //@}
    
    void finalize() {

      // Fill RATIO histograms (DataPointSets)
      vector<double> yvals_incl_ratio, yerrs_incl_ratio, yvals_excl_ratio, yerrs_excl_ratio,
        yvals_excl_pt150_ratio, yerrs_excl_pt150_ratio, yvals_excl_vbf_ratio, yerrs_excl_vbf_ratio;
      
      for (size_t i=0; i<6; ++i) {
        yvals_incl_ratio.push_back(ratio(_weights_incl[i+1], _weights_incl[i]));
        yvals_excl_ratio.push_back(ratio(_weights_excl[i+1], _weights_excl[i]));
        if (i>=1) yvals_excl_pt150_ratio.push_back(ratio(_weights_excl_pt150[i+1], _weights_excl_pt150[i]));
        if (i>=2) yvals_excl_vbf_ratio.push_back(ratio(_weights_excl_vbf[i+1], _weights_excl_vbf[i]));

        yerrs_incl_ratio.push_back(ratio_err_incl(_weights_incl[i+1], _weights_incl[i]));
        yerrs_excl_ratio.push_back(ratio_err_excl(_weights_excl[i+1], _weights_excl[i]));
        if (i>=1) yerrs_excl_pt150_ratio.push_back(ratio_err_excl(_weights_excl_pt150[i+1], _weights_excl_pt150[i]));
        if (i>=2) yerrs_excl_vbf_ratio.push_back(ratio_err_excl(_weights_excl_vbf[i+1], _weights_excl_vbf[i]));
      }


      // Actually fill histo
      _h_njet_incl_ratio  ->setCoordinate(1, yvals_incl_ratio,   yerrs_incl_ratio);
      _h_njet_excl_ratio  ->setCoordinate(1, yvals_excl_ratio,   yerrs_excl_ratio);
      _h_njet_excl_pt150_ratio  ->setCoordinate(1, yvals_excl_pt150_ratio,   yerrs_excl_pt150_ratio);
      _h_njet_excl_vbf_ratio  ->setCoordinate(1, yvals_excl_vbf_ratio,   yerrs_excl_vbf_ratio);
    
      
      double xs = crossSectionPerEvent()/picobarn;
      scale(_h_njet_incl      , xs);
      scale(_h_njet_excl      , xs);
      scale(_h_njet_excl_pt150, xs);
      scale(_h_njet_excl_vbf  , xs);
      scale(_h_ptlead         , xs);
      scale(_h_ptseclead      , xs);
      scale(_h_ptthirdlead    , xs);
      scale(_h_ptfourthlead   , xs);
      scale(_h_ptlead_excl    , xs);
      scale(_h_pt_ratio       , xs);
      scale(_h_pt_z           , xs);
      scale(_h_pt_z_excl      , xs);
      scale(_h_ylead          , xs);
      scale(_h_yseclead       , xs);
      scale(_h_ythirdlead     , xs);
      scale(_h_yfourthlead    , xs);
      scale(_h_deltay         , xs);
      scale(_h_mass           , xs);
      scale(_h_deltaphi       , xs);
      scale(_h_deltaR         , xs);
      scale(_h_ptthirdlead_vbf, xs);
      scale(_h_ythirdlead_vbf , xs);
      scale(_h_ht             , xs);
      scale(_h_st             , xs);
    }
      
    //@}


  protected:

    size_t _mode;

  private:

    std::vector<double> _weights_incl;
    std::vector<double> _weights_excl;
    std::vector<double> _weights_excl_pt150;
    std::vector<double> _weights_excl_vbf;

    //
    AIDA::IDataPointSet *_h_njet_incl_ratio;
    AIDA::IDataPointSet *_h_njet_excl_ratio;
    AIDA::IDataPointSet *_h_njet_excl_pt150_ratio;
    AIDA::IDataPointSet *_h_njet_excl_vbf_ratio;
    AIDA::IHistogram1D  *_h_njet_incl;
    AIDA::IHistogram1D  *_h_njet_excl;
    AIDA::IHistogram1D  *_h_njet_excl_pt150;
    AIDA::IHistogram1D  *_h_njet_excl_vbf;
    AIDA::IHistogram1D  *_h_ptlead;
    AIDA::IHistogram1D  *_h_ptseclead;
    AIDA::IHistogram1D  *_h_ptthirdlead;
    AIDA::IHistogram1D  *_h_ptfourthlead;
    AIDA::IHistogram1D  *_h_ptlead_excl;
    AIDA::IHistogram1D  *_h_pt_ratio;
    AIDA::IHistogram1D  *_h_pt_z;
    AIDA::IHistogram1D  *_h_pt_z_excl;
    AIDA::IHistogram1D  *_h_ylead;
    AIDA::IHistogram1D  *_h_yseclead;
    AIDA::IHistogram1D  *_h_ythirdlead;
    AIDA::IHistogram1D  *_h_yfourthlead;
    AIDA::IHistogram1D  *_h_deltay;
    AIDA::IHistogram1D  *_h_mass;
    AIDA::IHistogram1D  *_h_deltaphi;
    AIDA::IHistogram1D  *_h_deltaR;
    AIDA::IHistogram1D  *_h_ptthirdlead_vbf;
    AIDA::IHistogram1D  *_h_ythirdlead_vbf; 
    AIDA::IHistogram1D  *_h_ht;
    AIDA::IHistogram1D  *_h_st;
  };



  class ATLAS_2013_I1230812_EL : public ATLAS_2013_I1230812 {
  public:
    ATLAS_2013_I1230812_EL()
      : ATLAS_2013_I1230812("ATLAS_2013_I1230812_EL")
    {
      _mode = 2;
    }
  };



  class ATLAS_2013_I1230812_MU : public ATLAS_2013_I1230812 {
  public:
    ATLAS_2013_I1230812_MU()
      : ATLAS_2013_I1230812("ATLAS_2013_I1230812_MU")
    {
      _mode = 3;
    }
  };



  DECLARE_RIVET_PLUGIN(ATLAS_2013_I1230812);
  DECLARE_RIVET_PLUGIN(ATLAS_2013_I1230812_EL);
  DECLARE_RIVET_PLUGIN(ATLAS_2013_I1230812_MU);
}
